import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class GradientMethod(Algorithm):
    r"""
    Gradient method.

    See :doc:`3. Algorithm representation </theory/algorithm_representation>`
    for mathematical notation and definitions.

    Notation-driven assumptions are declared by the user via
    :class:`~autolyap.problemclass.InclusionProblem`: when present, terms written with
    :math:`\nabla` use differentiable functions, terms written with
    :math:`\prox_{\gamma f}` use proper, lower semicontinuous, convex functions,
    and terms written with :math:`J_{\gamma G}` use maximally monotone operators.

    Standard form
    -------------


    For an initial point :math:`x^0 \in \calH` and step size :math:`\gamma \in \reals_{++}`,

    .. math::
        (\forall k \in \naturals)\quad x^{k+1} = x^k - \gamma \nabla f(x^k).

    State-space representation
    --------------------------

    The update can be written in the algorithm representation with

    .. math::
        \bx^k = x^k, \qquad
        \bu^k = \nabla f(x^k), \qquad
        \by^k = x^k.

    With this representation, the system matrices are

    .. math::
        \begin{aligned}
            A_k &= \begin{bmatrix} 1 \end{bmatrix}, & B_k &= \begin{bmatrix} -\gamma \end{bmatrix}, \\
            C_k &= \begin{bmatrix} 1 \end{bmatrix}, & D_k &= \begin{bmatrix} 0 \end{bmatrix}.
        \end{aligned}

    These are the system matrices returned by :meth:`~autolyap.algorithms.Algorithm.get_ABCD`.

    Structural parameters
    ---------------------

    .. math::
        n = 1,\quad m = 1,\quad (\bar{m}_i)_{i=1}^{m} = (1),\quad \bar{m} = 1.

    .. math::
        I_{\text{func}} = \{1\},\quad I_{\text{op}} = \varnothing.
    """
    def __init__(self, gamma: float) -> None:
        r"""
        Initialize the gradient method.

        Structural inputs passed to :class:`~autolyap.algorithms.Algorithm` are

        .. math::
            n = 1,\quad m = 1,\quad (\bar m_i)_{i=1}^{m} = (1),\quad \bar m = 1,\quad
            I_{\mathrm{func}} = \{1\},\quad I_{\mathrm{op}} = \varnothing.
        """
        super().__init__(n=1, m=1, m_bar_is=[1], I_func=[1], I_op=[])
        self.gamma = gamma
    
    def set_gamma(self, gamma: float) -> None:
        r"""
        Set the step-size parameter :math:`\gamma`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.GradientMethod`.

        **Parameters**

        - `gamma` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`\gamma`.

        **Raises**

        - `ValueError`: If `gamma` is not a finite real number or if :math:`\gamma \le 0`.
        """
        gamma = self._validate_positive_finite_real(gamma, "gamma")
        self._set_dynamic_parameter("gamma", gamma)
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1]])
        B = np.array([[-self.gamma]])
        C = np.array([[1]])
        D = np.array([[0]])
        return (A, B, C, D)



from .helper_functions import create_symmetric_matrix_expression
from .helper_functions import create_symmetric_matrix
from .validation import (
    ensure_finite_array,
    ensure_index_list,
    ensure_integral,
    ensure_m_bar_list,
    ensure_real_number,
)

__all__ = [
    'create_symmetric_matrix_expression',
    'create_symmetric_matrix',
    'ensure_finite_array',
    'ensure_index_list',
    'ensure_integral',
    'ensure_m_bar_list',
    'ensure_real_number',
]



import numpy as np
import math
from typing import Tuple
from .algorithm import Algorithm

class ITEM(Algorithm):
    r"""
    Information-theoretic exact method (ITEM) :cite:`taylor2023optimalgradientmethod`.

    See :doc:`3. Algorithm representation </theory/algorithm_representation>`
    for mathematical notation and definitions.

    Notation-driven assumptions are declared by the user via
    :class:`~autolyap.problemclass.InclusionProblem`: when present, terms written with
    :math:`\nabla` use differentiable functions, terms written with
    :math:`\prox_{\gamma f}` use proper, lower semicontinuous, convex functions,
    and terms written with :math:`J_{\gamma G}` use maximally monotone operators.

    Standard form
    -------------


    Let :math:`q = \mu/L`, :math:`\tilde{A}_0 = 0`, and for all :math:`k \in \naturals`,

    .. math::
        \begin{aligned}
            \tilde{A}_{k+1}
            &= \frac{(1+q)\tilde{A}_k + 2\left(1 + \sqrt{(1+\tilde{A}_k)(1+q\tilde{A}_k)}\right)}{(1-q)^2}, \\
            \beta_k
            &= \frac{\tilde{A}_k}{(1-q)\tilde{A}_{k+1}}, \\
            \delta_k
            &= \frac{(1-q)^2\tilde{A}_{k+1} - (1+q)\tilde{A}_k}{2(1+q+q\tilde{A}_k)}.
        \end{aligned}

    For initial points :math:`x^0,z^0 \in \calH`,

    .. math::
        (\forall k \in \naturals)\quad
        \left[
        \begin{aligned}
            y^k &= (1-\beta_k)z^k + \beta_k x^k, \\
            x^{k+1} &= y^k - \frac{1}{L}\nabla f(y^k), \\
            z^{k+1} &= (1-q\delta_k)z^k + q\delta_k y^k - \frac{\delta_k}{L}\nabla f(y^k).
        \end{aligned}
        \right.

    State-space representation
    --------------------------

    The update can be written in the algorithm representation with

    .. math::
        \bx^k = (x^k, z^k), \qquad
        \bu^k = \nabla f(y^k), \qquad
        \by^k = y^k.

    Let :math:`q = \mu/L`, with :math:`\beta_k` and :math:`\delta_k` as in the
    standard form above. The system matrices are

    .. math::
        \begin{aligned}
            A_k &=
            \begin{bmatrix}
            \beta_k & 1-\beta_k \\
            q\beta_k\delta_k & 1-q\beta_k\delta_k
            \end{bmatrix}, &
            B_k &=
            \begin{bmatrix}
            -\frac{1}{L} \\
            -\frac{\delta_k}{L}
            \end{bmatrix}, \\
            C_k &= \begin{bmatrix} \beta_k & 1-\beta_k \end{bmatrix}, &
            D_k &= \begin{bmatrix} 0 \end{bmatrix}.
        \end{aligned}

    These are the system matrices returned by :meth:`~autolyap.algorithms.Algorithm.get_ABCD`.

    Structural parameters
    ---------------------

    .. math::
        n = 2,\quad m = 1,\quad (\bar{m}_i)_{i=1}^{m} = (1),\quad \bar{m} = 1.

    .. math::
        I_{\text{func}} = \{1\},\quad I_{\text{op}} = \varnothing.
    """
    def __init__(self, mu: float, L: float) -> None:
        r"""
        Initialize ITEM.

        Structural inputs passed to :class:`~autolyap.algorithms.Algorithm` are

        .. math::
            n = 2,\quad m = 1,\quad (\bar m_i)_{i=1}^{m} = (1),\quad \bar m = 1,\quad
            I_{\mathrm{func}} = \{1\},\quad I_{\mathrm{op}} = \varnothing.
        """
        super().__init__(2, 1, [1], [1], [])
        self.set_L(L)
        self.set_mu(mu)

    @staticmethod
    def _validate_mu_lt_L(mu: float, L: float) -> None:
        if mu >= L:
            raise ValueError(f"Require mu < L. Got mu={mu}, L={L}.")
    
    def set_L(self, L: float) -> None:
        r"""
        Set the smoothness parameter :math:`L`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.ITEM`.

        **Parameters**

        - `L` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`L`.

        **Raises**

        - `ValueError`: If `L` is not a finite real number, if :math:`L \le 0`,
          or if :math:`\mu \ge L`.
        """
        L = self._validate_positive_finite_real(L, "L")
        if hasattr(self, "mu"):
            self._validate_mu_lt_L(self.mu, L)
        self._set_dynamic_parameter("L", L)

    def set_mu(self, mu: float) -> None:
        r"""
        Set the strong-convexity parameter :math:`\mu`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.ITEM`.

        **Parameters**

        - `mu` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`\mu`.

        **Raises**

        - `ValueError`: If `mu` is not a finite real number, if :math:`\mu \le 0`,
          or if :math:`\mu \ge L`.
        """
        mu = self._validate_positive_finite_real(mu, "mu")
        if hasattr(self, "L"):
            self._validate_mu_lt_L(mu, self.L)
        self._set_dynamic_parameter("mu", mu)
    
    def get_A(self, k: int) -> float:
        r"""
        Return the scalar sequence value :math:`\tilde{A}_k`.

        This scalar follows the recurrence stated in the class-level reference of
        :class:`~autolyap.algorithms.ITEM`.
        The system matrix :math:`A_k` in :meth:`get_ABCD` uses the same letter by
        convention; this accessor returns only the scalar sequence
        :math:`\tilde{A}_k`.
        """
        k = self._validate_nonnegative_integral(k, "k")
        q = self.mu / self.L
        A_tilde = 0.0
        for _ in range(k):
            A_tilde = (
                (1 + q) * A_tilde
                + 2 * (1 + math.sqrt((1 + A_tilde) * (1 + q * A_tilde)))
            ) / ((1 - q) ** 2)
        return A_tilde

    def compute_beta(self, k: int) -> float:
        k = self._validate_nonnegative_integral(k, "k")
        q = self.mu / self.L
        A_tilde_k = self.get_A(k)
        A_tilde_k1 = self.get_A(k + 1)
        return A_tilde_k / ((1 - q) * A_tilde_k1)

    def compute_delta(self, k: int) -> float:
        k = self._validate_nonnegative_integral(k, "k")
        q = self.mu / self.L
        A_tilde_k = self.get_A(k)
        A_tilde_k1 = self.get_A(k + 1)
        numerator = ((1 - q) ** 2) * A_tilde_k1 - (1 + q) * A_tilde_k
        denominator = 2 * (1 + q + q * A_tilde_k)
        return numerator / denominator

    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        k = self._validate_nonnegative_integral(k, "k")
        q = self.mu / self.L
        beta = self.compute_beta(k)
        delta = self.compute_delta(k)

        A = np.array([[beta, 1-beta],
                      [q*beta*delta, 1-q*beta*delta]])
        
        B = np.array([[-1/self.L],
                      [-delta/self.L]])
        
        C = np.array([[beta, 1-beta]])
        
        D = np.array([[0]])
        
        return (A, B, C, D)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class NesterovFastGradientMethod(Algorithm):
    r"""
    Nesterov's fast gradient method :cite:`nesterov1983fast`.

    See :doc:`3. Algorithm representation </theory/algorithm_representation>`
    for mathematical notation and definitions.

    Notation-driven assumptions are declared by the user via
    :class:`~autolyap.problemclass.InclusionProblem`: when present, terms written with
    :math:`\nabla` use differentiable functions, terms written with
    :math:`\prox_{\gamma f}` use proper, lower semicontinuous, convex functions,
    and terms written with :math:`J_{\gamma G}` use maximally monotone operators.

    Standard form
    -------------


    For initial points :math:`x^{-1},x^0 \in \calH`, step size :math:`\gamma \in \reals_{++}`,
    and :math:`\lambda_0 = 1`,

    .. math::
        (\forall k \in \naturals)\quad
        \left[
        \begin{aligned}
            y^k &= x^k + \delta_k (x^k - x^{k-1}), \\
            x^{k+1} &= y^k - \gamma \nabla f(y^k), \\
            \lambda_{k+1} &= \frac{1 + \sqrt{1 + 4\lambda_k^2}}{2}, \\
            \delta_k &= \frac{\lambda_k - 1}{\lambda_{k+1}}.
        \end{aligned}
        \right.

    This implementation also keeps an additional evaluation of :math:`\nabla f(x^k)`
    for analysis templates that require it.

    State-space representation
    --------------------------

    The update can be written in the algorithm representation with

    .. math::
        \bx^k = (x^k, x^{k-1}), \qquad
        \bu^k = (\nabla f(y^k), \nabla f(x^k)), \qquad
        \by^k = (y^k, x^k).

    With :math:`\lambda_0 = 1`, :math:`\lambda_{k+1} = \frac{1+\sqrt{1+4\lambda_k^2}}{2}`,
    and :math:`\alpha_k = \frac{\lambda_k - 1}{\lambda_{k+1}}`, the system
    matrices are

    .. math::
        \begin{aligned}
            A_k &=
            \begin{bmatrix}
            1+\alpha_k & -\alpha_k \\
            1 & 0
            \end{bmatrix}, &
            B_k &=
            \begin{bmatrix}
            -\gamma & 0 \\
            0 & 0
            \end{bmatrix}, \\
            C_k &=
            \begin{bmatrix}
            1+\alpha_k & -\alpha_k \\
            1 & 0
            \end{bmatrix}, &
            D_k &=
            \begin{bmatrix}
            0 & 0 \\
            0 & 0
            \end{bmatrix}.
        \end{aligned}

    These are the system matrices returned by :meth:`~autolyap.algorithms.Algorithm.get_ABCD`.

    Structural parameters
    ---------------------

    .. math::
        n = 2,\quad m = 1,\quad (\bar{m}_i)_{i=1}^{m} = (2),\quad \bar{m} = 2.

    .. math::
        I_{\text{func}} = \{1\},\quad I_{\text{op}} = \varnothing.
    """
    def __init__(self, gamma: float) -> None:
        r"""
        Initialize the fast gradient method.

        Structural inputs passed to :class:`~autolyap.algorithms.Algorithm` are

        .. math::
            n = 2,\quad m = 1,\quad (\bar m_i)_{i=1}^{m} = (2),\quad \bar m = 2,\quad
            I_{\mathrm{func}} = \{1\},\quad I_{\mathrm{op}} = \varnothing.
        """
        super().__init__(2, 1, [2], [1], [])
        self.gamma = gamma
    
    def set_gamma(self, gamma: float) -> None:
        r"""
        Set the step-size parameter :math:`\gamma`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.NesterovFastGradientMethod`.

        **Parameters**

        - `gamma` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`\gamma`.

        **Raises**

        - `ValueError`: If `gamma` is not a finite real number or if :math:`\gamma \le 0`.
        """
        gamma = self._validate_positive_finite_real(gamma, "gamma")
        self._set_dynamic_parameter("gamma", gamma)
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        lambda_var = 1
        for _ in range(0, k + 1):
            lambda_var_prev = lambda_var
            lambda_var = (1 + np.sqrt(1 + 4 * lambda_var ** 2)) / 2

        alpha = (lambda_var_prev - 1) / lambda_var

        A = np.array([[1+alpha, -alpha],
                      [1, 0]])
        
        B = np.array([[-self.gamma, 0],
                      [0, 0]])
        
        C = np.array([[1+alpha, -alpha],
                      [1, 0]])
        
        D = np.array([[0, 0],
                      [0, 0]])
        
        return (A, B, C, D)



"""Concrete function interpolation-condition classes."""

import numpy as np
from typing import List, Union, Tuple

from autolyap.problemclass.base import _FunctionInterpolationCondition
from autolyap.problemclass.indices import _InterpolationIndices
from autolyap.utils.validation import ensure_real_number

INF = float("inf")


def _ensure_positive_finite(value: Union[int, float], parameter_name: str, error_message: str) -> float:
    r"""Return `value` as float after enforcing strict positivity and finiteness."""
    numeric_value = ensure_real_number(value, parameter_name)
    if not np.isfinite(numeric_value) or numeric_value <= 0:
        raise ValueError(error_message)
    return numeric_value


def _ensure_positive_mu_tilde(mu_tilde: Union[int, float], context_name: str) -> Tuple[float, float]:
    r"""
    Validate :math:`\tilde{\mu} > 0` and return both :math:`(\tilde{\mu}, -\tilde{\mu})`.

    Many weakly-convex templates are written with :math:`\mu=-\tilde{\mu}`.
    This helper keeps that conversion explicit and centralized.
    """
    validated = _ensure_positive_finite(
        mu_tilde,
        "Parameter mu_tilde",
        f"For {context_name}, mu_tilde must be > 0 and finite.",
    )
    return validated, -validated


class _ParametrizedFunctionInterpolationCondition(_FunctionInterpolationCondition):
    r"""
    Base class for function interpolation conditions parameterized by :math:`\mu` and :math:`L`.

    Class-level reference
    =====================

    This class-level docstring centralizes notation for the shared
    :math:`(\mu, L)` interpolation family used by its concrete subclasses.

    Provides a helper to compute interpolation data from :math:`\mu` and :math:`L`.
    This base class supports both smooth and nonsmooth conditions by setting :math:`L` appropriately.

    This template applies to :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` with
    parameters satisfying :math:`-\infty < \mu < L \le +\infty` and :math:`L > 0`
    (nonsmooth cases are encoded by :math:`L = +\infty`).

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} \in \partial f(x_{r_1})`, :math:`g_{r_2} \in \partial f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`, the condition enforced is

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle + \frac{\mu}{2}\|x_{r_1} - x_{r_2}\|^2
          + \frac{1}{2(L-\mu)}\|g_{r_1} - g_{r_2} - \mu(x_{r_1} - x_{r_2})\|^2,

      where :math:`\frac{1}{2(L-\mu)}` is interpreted as :math:`0` when :math:`L = +\infty`.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`F = (F_{r_1}, F_{r_2})` and
      :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})`, the interpolation inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0, \qquad a = (-1, 1),

      with

      .. math::
          M =
          \begin{cases}
            \frac{1}{2(L-\mu)}
            \begin{bmatrix}
              L\mu & -L\mu & -\mu & L \\
              -L\mu & L\mu & \mu & -L \\
              -\mu & \mu & 1 & -1 \\
              L & -L & -1 & 1
            \end{bmatrix}, & \text{if } L < +\infty, \\
            \frac{1}{2}
            \begin{bmatrix}
              \mu & -\mu & 0 & 1 \\
              -\mu & \mu & 0 & -1 \\
              0 & 0 & 0 & 0 \\
              1 & -1 & 0 & 0
            \end{bmatrix}, & \text{if } L = +\infty.
          \end{cases}

      The returned interpolation indices are ``r1!=r2``.
      The `eq` flag returned by :meth:`get_data` is ``False``.

    Note: Many classes below are specializations obtained by choosing :math:`\mu` and :math:`L`
    (e.g., weakly convex uses :math:`\mu = -\tilde{\mu}` with :math:`L=+\infty`; smooth uses
    :math:`\mu = -L`).

    **Parameters**

    - `mu` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The convexity parameter corresponding to :math:`\mu`.
      For strongly convex functions, :math:`\mu > 0`; for convex functions, :math:`\mu = 0`; for weakly convex
      functions, :math:`\mu < 0`.
    - `L` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The smoothness parameter corresponding to :math:`L`.
      For nonsmooth functions, :math:`L` is set to infinity.

    **Raises**

    - `ValueError`: If `mu` is not a number, if :math:`L \le 0` with :math:`L < +\infty`,
      or if :math:`\mu \geq L`.
    """
    def __init__(self, mu: Union[int, float], L: Union[int, float]) -> None:
        mu, L = self._validate_mu_L(mu, L)
        self.mu = mu
        self.L = L
        self._interpolation_data = self._compute_interpolation_data(mu, L)

    @staticmethod
    def _validate_mu_L(mu: Union[int, float], L: Union[int, float]) -> Tuple[float, float]:
        r"""
        Validate interpolation parameters and return normalized ``(mu, L)``.

        Enforces the admissible region ``-inf < mu < L <= +inf`` with the
        additional requirement that finite ``L`` values are strictly positive.
        """
        mu = ensure_real_number(mu, "Parameter mu")
        L = ensure_real_number(L, "Parameter L")
        if L != INF and L <= 0:
            raise ValueError("Parameter L must be positive or +inf.")
        if mu == -INF:
            raise ValueError("ParametrizedFunctionInterpolationCondition: mu cannot be -inf.")
        if not (mu < L):
            raise ValueError("ParametrizedFunctionInterpolationCondition requires that -inf < mu < L <= +inf.")
        return mu, L

    @staticmethod
    def _compute_interpolation_data(mu: float, L: float) -> Tuple[np.ndarray, np.ndarray, bool, _InterpolationIndices]:
        r"""
        Compute the interpolation data based on mu and L.

        The tuple format and notation follow the class-level reference in
        :class:`~autolyap.problemclass.functions._ParametrizedFunctionInterpolationCondition`.

        When L is infinite, the condition is nonsmooth and a simpler interpolation matrix is used.
        When L is finite, the interpolation data follows the formula for smooth functions.

        **Parameters**

        - `mu` (:class:`float`): The convexity parameter corresponding to :math:`\mu`.
        - `L` (:class:`float`): The smoothness parameter corresponding to :math:`L`.

        **Returns**

        - (:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`numpy.ndarray`, :class:`bool`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]): A tuple containing the
          matrix, vector, eq flag, and interpolation indices.

        """
        if L == INF:
            matrix = 0.5 * np.array([
                [mu, -mu, 0, 1],
                [-mu, mu, 0, -1],
                [0, 0, 0, 0],
                [1, -1, 0, 0]
            ])
        else:
            matrix = (1 / (2 * (L - mu))) * np.array([
                [L * mu, -L * mu, -mu, L],
                [-L * mu, L * mu, mu, -L],
                [-mu, mu, 1, -1],
                [L, -L, -1, 1]
            ])
        vector = np.array([-1, 1])
        eq = False
        interp_idx = _InterpolationIndices("r1!=r2")
        return matrix, vector, eq, interp_idx

    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, _InterpolationIndices]]:
        r"""
        Return interpolation data for the function condition.

        The tuple format and notation follow the class-level reference in
        :class:`~autolyap.problemclass.functions._ParametrizedFunctionInterpolationCondition`.

        **Returns**

        - (:class:`~typing.List`\[:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`numpy.ndarray`, :class:`bool`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]\]): A list containing one
          tuple with the matrix, vector, eq flag, and interpolation indices.

        """
        return [self._interpolation_data]

class Convex(_ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for proper, lower semicontinuous, and convex functions.

    Let :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be proper, lower semicontinuous, and convex.

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} \in \partial f(x_{r_1})`, :math:`g_{r_2} \in \partial f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`,

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})` and
      :math:`F = (F_{r_1}, F_{r_2})`, the same inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0,

      with

      .. math::
          a = (-1, 1), \qquad
          M = \frac{1}{2}
          \begin{bmatrix}
              0 & 0 & 0 & 1 \\
              0 & 0 & 0 & -1 \\
              0 & 0 & 0 & 0 \\
              1 & -1 & 0 & 0
          \end{bmatrix}.

    **References**

    - :cite:`taylor2016smoothstronglyconvex{Theorem 4}`.

    """
    def __init__(self) -> None:
        super().__init__(mu=0.0, L=INF)

class StronglyConvex(_ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for proper, lower semicontinuous, and strongly convex functions.

    Let :math:`\mu \in \mathbb{R}_{++}` and
    :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be proper, lower semicontinuous,
    and :math:`\mu`-strongly convex.

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} \in \partial f(x_{r_1})`, :math:`g_{r_2} \in \partial f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`,

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle + \frac{\mu}{2}\|x_{r_1} - x_{r_2}\|^2.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})` and
      :math:`F = (F_{r_1}, F_{r_2})`, the same inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0,

      with

      .. math::
          a = (-1, 1), \qquad
          M = \frac{1}{2}
          \begin{bmatrix}
              \mu & -\mu & 0 & 1 \\
              -\mu & \mu & 0 & -1 \\
              0 & 0 & 0 & 0 \\
              1 & -1 & 0 & 0
          \end{bmatrix}.

    **Parameters**

    - `mu` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Strong convexity parameter corresponding to
      :math:`\mu` (must be :math:`> 0` and finite).

    **Raises**

    - `ValueError`: If `mu` is not valid.

    **References**

    - :cite:`taylor2016smoothstronglyconvex{Theorem 4}`.
    """
    def __init__(self, mu: Union[int, float]) -> None:
        mu = _ensure_positive_finite(mu, "Parameter mu", "For StronglyConvex, mu must be > 0 and finite.")
        super().__init__(mu=mu, L=INF)

class WeaklyConvex(_ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for proper, lower semicontinuous, and weakly convex functions.

    Let :math:`\tilde{\mu} \in \mathbb{R}_{++}` and
    :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be proper, lower semicontinuous,
    and :math:`\tilde{\mu}`-weakly convex.

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} \in \partial f(x_{r_1})`, :math:`g_{r_2} \in \partial f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`,

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle - \frac{\tilde{\mu}}{2}\|x_{r_1} - x_{r_2}\|^2.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})` and
      :math:`F = (F_{r_1}, F_{r_2})`, the same inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0,

      with

      .. math::
          a = (-1, 1), \qquad
          M = \frac{1}{2}
          \begin{bmatrix}
              -\tilde{\mu} & \tilde{\mu} & 0 & 1 \\
              \tilde{\mu} & -\tilde{\mu} & 0 & -1 \\
              0 & 0 & 0 & 0 \\
              1 & -1 & 0 & 0
          \end{bmatrix}.

    **Parameters**

    - `mu_tilde` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Weak convexity parameter corresponding to
      :math:`\tilde{\mu}` (must be :math:`> 0` and finite).

    **Raises**

    - `ValueError`: If `mu_tilde` is not valid.

    **References**

    - :cite:`rotaru2022tightconvergencerates{Theorem 3.1}`.
    """
    def __init__(self, mu_tilde: Union[int, float]) -> None:
        mu_tilde, mu = _ensure_positive_mu_tilde(mu_tilde, "WeaklyConvex")
        super().__init__(mu=mu, L=INF)
        self.mu_tilde = mu_tilde

class Smooth(_ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for smooth functions.

    Let :math:`L \in \mathbb{R}_{++}` and
    :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be Fréchet differentiable
    with :math:`L`-Lipschitz gradient.

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} = \nabla f(x_{r_1})`, :math:`g_{r_2} = \nabla f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`,

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle
          + \frac{1}{4L}\|g_{r_1} - g_{r_2} + L(x_{r_1} - x_{r_2})\|^2
          - \frac{L}{2}\|x_{r_1} - x_{r_2}\|^2.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})` and
      :math:`F = (F_{r_1}, F_{r_2})`, the same inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0,

      with

      .. math::
          a = (-1, 1), \qquad
          M = \frac{1}{4L}
          \begin{bmatrix}
              -L^2 & L^2 & L & L \\
              L^2 & -L^2 & -L & -L \\
              L & -L & 1 & -1 \\
              L & -L & -1 & 1
          \end{bmatrix}.

    **Parameters**

    - `L` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Smoothness parameter corresponding to
      :math:`L` (must be :math:`> 0` and finite).

    **Raises**

    - `ValueError`: If `L` is not valid.

    **References**

    - :cite:`taylor2017exactworstcase{Theorem 3.10}`.
    """
    def __init__(self, L: Union[int, float]) -> None:
        L = _ensure_positive_finite(L, "Parameter L", "For Smooth, L must be > 0 and finite.")
        super().__init__(mu=-L, L=L)

class SmoothConvex(_ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for smooth and convex functions.

    Let :math:`L \in \mathbb{R}_{++}` and
    :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be convex, Fréchet differentiable,
    with :math:`L`-Lipschitz gradient.

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} = \nabla f(x_{r_1})`, :math:`g_{r_2} = \nabla f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`,

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle
          + \frac{1}{2L}\|g_{r_1} - g_{r_2}\|^2.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})` and
      :math:`F = (F_{r_1}, F_{r_2})`, the same inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0,

      with

      .. math::
          a = (-1, 1), \qquad
          M = \frac{1}{2L}
          \begin{bmatrix}
              0 & 0 & 0 & L \\
              0 & 0 & 0 & -L \\
              0 & 0 & 1 & -1 \\
              L & -L & -1 & 1
          \end{bmatrix}.

    **Parameters**

    - `L` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Smoothness parameter corresponding to
      :math:`L` (must be :math:`> 0` and finite).

    **Raises**

    - `ValueError`: If `L` is not valid.

    **References**

    - :cite:`taylor2016smoothstronglyconvex{Theorem 4}`.
    """
    def __init__(self, L: Union[int, float]) -> None:
        L = _ensure_positive_finite(L, "Parameter L", "For SmoothConvex, L must be > 0 and finite.")
        super().__init__(mu=0.0, L=L)

class SmoothStronglyConvex(_ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for smooth and strongly convex functions.

    Let :math:`0 < \mu < L` and
    :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be :math:`\mu`-strongly convex,
    Fréchet differentiable, with :math:`L`-Lipschitz gradient.

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} = \nabla f(x_{r_1})`, :math:`g_{r_2} = \nabla f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`,

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle + \frac{\mu}{2}\|x_{r_1} - x_{r_2}\|^2
          + \frac{1}{2(L-\mu)}\|g_{r_1} - g_{r_2} - \mu(x_{r_1} - x_{r_2})\|^2.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})` and
      :math:`F = (F_{r_1}, F_{r_2})`, the same inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0,

      with

      .. math::
          a = (-1, 1), \qquad
          M = \frac{1}{2(L-\mu)}
          \begin{bmatrix}
              L\mu & -L\mu & -\mu & L \\
              -L\mu & L\mu & \mu & -L \\
              -\mu & \mu & 1 & -1 \\
              L & -L & -1 & 1
          \end{bmatrix}.

    **Parameters**

    - `mu` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Strong convexity parameter corresponding to
      :math:`\mu` (must be :math:`> 0` and finite).
    - `L` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Smoothness parameter corresponding to :math:`L`
      (must be :math:`> 0` and finite) with :math:`\mu < L`.

    **Raises**

    - `ValueError`: If parameters are not valid.

    **References**

    - :cite:`taylor2016smoothstronglyconvex{Theorem 4}`.
    """
    def __init__(self, mu: Union[int, float], L: Union[int, float]) -> None:
        mu = _ensure_positive_finite(
            mu,
            "Parameter mu",
            "For SmoothStronglyConvex, mu must be > 0 and finite.",
        )
        L = _ensure_positive_finite(
            L,
            "Parameter L",
            "For SmoothStronglyConvex, L must be > 0 and finite.",
        )
        if mu >= L:
            raise ValueError("For SmoothStronglyConvex, mu must be less than L.")
        super().__init__(mu=mu, L=L)

class SmoothWeaklyConvex(_ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for smooth and weakly convex functions.

    Let :math:`\tilde{\mu} \in \mathbb{R}_{++}` and :math:`L \in \mathbb{R}_{++}`, with
    :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be :math:`\tilde{\mu}`-weakly convex,
    Fréchet differentiable, with :math:`L`-Lipschitz gradient.

    - Interpolation inequality:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} = \nabla f(x_{r_1})`, :math:`g_{r_2} = \nabla f(x_{r_2})`,
      and :math:`F_{r_1} = f(x_{r_1})`, :math:`F_{r_2} = f(x_{r_2})`,

      .. math::
          F_{r_1} \ge F_{r_2} + \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle - \frac{\tilde{\mu}}{2}\|x_{r_1} - x_{r_2}\|^2
          + \frac{1}{2(L+\tilde{\mu})}\|g_{r_1} - g_{r_2} + \tilde{\mu}(x_{r_1} - x_{r_2})\|^2.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})` and
      :math:`F = (F_{r_1}, F_{r_2})`, the same inequality is encoded as

      .. math::
          a^\top F + \mathcal{Q}\p{M, z} \le 0,

      with

      .. math::
          a = (-1, 1), \qquad
          M = \frac{1}{2(L+\tilde{\mu})}
          \begin{bmatrix}
              -L\tilde{\mu} & L\tilde{\mu} & \tilde{\mu} & L \\
              L\tilde{\mu} & -L\tilde{\mu} & -\tilde{\mu} & -L \\
              \tilde{\mu} & -\tilde{\mu} & 1 & -1 \\
              L & -L & -1 & 1
          \end{bmatrix}.

    **Parameters**

    - `mu_tilde` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Weak convexity parameter corresponding to
      :math:`\tilde{\mu}` (must be :math:`> 0` and finite).
    - `L` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): Smoothness parameter corresponding to
      :math:`L` (must be :math:`> 0` and finite).

    **Raises**

    - `ValueError`: If parameters are not valid.

    **References**

    - :cite:`rotaru2022tightconvergencerates{Theorem 3.1}`.
    """
    def __init__(self, mu_tilde: Union[int, float], L: Union[int, float]) -> None:
        mu_tilde, mu = _ensure_positive_mu_tilde(mu_tilde, "SmoothWeaklyConvex")
        L = _ensure_positive_finite(
            L,
            "Parameter L",
            "For SmoothWeaklyConvex, L must be > 0 and finite.",
        )
        super().__init__(mu=mu, L=L)
        self.mu_tilde = mu_tilde

class IndicatorFunctionOfClosedConvexSet(_FunctionInterpolationCondition):
    r"""
    Function interpolation condition for indicator functions of nonempty, closed, and convex sets.

    Let :math:`C \subseteq \calH` be nonempty, closed, and convex, and define
    its indicator function :math:`\delta_C : \calH \to \reals \cup \{\pm\infty\}`
    by

    .. math::
        \delta_C(x) =
        \begin{cases}
            0, & \text{if } x \in C, \\
            +\infty, & \text{if } x \notin C.
        \end{cases}

    Let :math:`N_C:\calH\rightrightarrows\calH` denote the normal cone of :math:`C`, defined by

    .. math::
        N_C(x) =
        \begin{cases}
            \{g \in \calH : \langle g, z - x \rangle \le 0,\ \forall z \in C\},
            & \text{if } x \in C, \\
            \emptyset, & \text{if } x \notin C,
        \end{cases}

    which coincides with the subdifferential of the indicator:

    .. math::
        \partial \delta_C(x) = N_C(x), \qquad \forall x \in \calH.

    - Interpolation inequalities used:

      For any :math:`x_{r_1}, x_{r_2} \in C` with
      :math:`g_{r_1} \in N_C(x_{r_1})`, :math:`g_{r_2} \in N_C(x_{r_2})`,
      and :math:`F_{r_1} = \delta_C(x_{r_1})`, :math:`F_{r_2} = \delta_C(x_{r_2})`,

      .. math::
          \langle g_{r_2}, x_{r_1} - x_{r_2} \rangle \le 0 \quad \text{for } r_1 \ne r_2,

      and

      .. math::
          F_{r_1} = 0.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`F = (F_{r_1}, F_{r_2})` and
      :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})`, the inequality
      constraint has

      .. math::
          a_1 = (0, 0), \qquad
          M_1 = \frac{1}{2}
          \begin{bmatrix}
              0 & 0 & 0 & 1 \\
              0 & 0 & 0 & -1 \\
              0 & 0 & 0 & 0 \\
              1 & -1 & 0 & 0
          \end{bmatrix}.

      For the equality :math:`F_{r_1}=0`, with
      :math:`\hat{z}=(x_{r_1}, g_{r_1})`, the coefficients are

      .. math::
          a_2 = (1), \qquad
          M_2 =
          \begin{bmatrix}
              0 & 0 \\
              0 & 0
          \end{bmatrix}.

    This condition has no parameters.

    **References**

    - :cite:`taylor2017exactworstcase{Theorem 3.6}`.

    """
    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, _InterpolationIndices]]:
        r"""
        Return interpolation data for the indicator function.

        The tuple format and notation follow the class-level reference in
        :class:`~autolyap.problemclass.IndicatorFunctionOfClosedConvexSet`.

        **Returns**

        - (:class:`~typing.List`\[:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`numpy.ndarray`, :class:`bool`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]\]): A list containing two
          tuples with the interpolation data.

        """
        interp_idx_ineq = _InterpolationIndices("r1!=r2")
        matrix_ineq = 0.5 * np.array([
            [0, 0, 0, 1],
            [0, 0, 0, -1],
            [0, 0, 0, 0],
            [1, -1, 0, 0]
        ])
        vector_ineq = np.array([0, 0])
        interp_idx_eq = _InterpolationIndices("r1")
        matrix_eq = np.array([[0, 0], [0, 0]])
        vector_eq = np.array([1])
        return [
            (matrix_ineq, vector_ineq, False, interp_idx_ineq),
            (matrix_eq, vector_eq, True, interp_idx_eq)
        ]

class SupportFunctionOfClosedConvexSet(_FunctionInterpolationCondition):
    r"""
    Function interpolation condition for support functions of nonempty, closed, and convex sets.

    Let :math:`C \subseteq \calH` be nonempty, closed, and convex, and define
    its support function :math:`\sigma_C : \calH \to \reals \cup \{\pm\infty\}`
    by

    .. math::
        \sigma_C(x) = \sup_{c \in C} \langle x, c \rangle.

    - Interpolation inequalities used:

      For any :math:`x_{r_1}, x_{r_2} \in \calH` with
      :math:`g_{r_1} \in \partial \sigma_C(x_{r_1})`, :math:`g_{r_2} \in \partial \sigma_C(x_{r_2})`,
      and :math:`F_{r_1} = \sigma_C(x_{r_1})`, :math:`F_{r_2} = \sigma_C(x_{r_2})`,

      .. math::
        F_{r_1} = \langle x_{r_1}, g_{r_1} \rangle,

      and

      .. math::
        \langle x_{r_2}, g_{r_1} - g_{r_2} \rangle \le 0.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`F = (F_{r_1}, F_{r_2})` and
      :math:`z = (x_{r_1}, x_{r_2}, g_{r_1}, g_{r_2})`, the inequality
      constraint has

      .. math::
          a_1 = (0, 0), \qquad
          M_1 = \frac{1}{2}
          \begin{bmatrix}
              0 & 0 & 0 & 0 \\
              0 & 0 & 1 & -1 \\
              0 & 1 & 0 & 0 \\
              0 & -1 & 0 & 0
          \end{bmatrix}.

      For the equality :math:`F_{r_1}=\langle x_{r_1},g_{r_1}\rangle`,
      with :math:`\hat{z}=(x_{r_1}, g_{r_1})`, the coefficients are

      .. math::
          a_2 = (-1), \qquad
          M_2 = \frac{1}{2}
          \begin{bmatrix}
              0 & 1 \\
              1 & 0
          \end{bmatrix}.

    This condition has no parameters.

    **References**

    - :cite:`taylor2017exactworstcase{Corollary 3.7}`.

    """
    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, _InterpolationIndices]]:
        r"""
        Return interpolation data for the support function.

        The tuple format and notation follow the class-level reference in
        :class:`~autolyap.problemclass.SupportFunctionOfClosedConvexSet`.

        **Returns**

        - (:class:`~typing.List`\[:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`numpy.ndarray`, :class:`bool`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]\]): A list containing two
          tuples with the interpolation data.

        """
        interp_idx_ineq = _InterpolationIndices("r1!=r2")
        matrix_ineq = 0.5 * np.array([
            [0, 0, 0, 0],
            [0, 0, 1, -1],
            [0, 1, 0, 0],
            [0, -1, 0, 0]
        ])
        vector_ineq = np.array([0, 0])
        interp_idx_eq = _InterpolationIndices("r1")
        matrix_eq = 0.5 * np.array([[0, 1], [1, 0]])
        vector_eq = np.array([-1])
        return [
            (matrix_ineq, vector_ineq, False, interp_idx_ineq),
            (matrix_eq, vector_eq, True, interp_idx_eq)
        ]

class GradientDominated(_FunctionInterpolationCondition):
    r"""
    Function interpolation condition for gradient-dominated functions.

    Let :math:`\mu_{\textup{gd}} \in \mathbb{R}_{++}` and
    :math:`f: \calH \to \mathbb{R} \cup \{\pm\infty\}` be Fréchet differentiable and
    :math:`\mu_{\textup{gd}}`-gradient dominated.

    - Interpolation inequalities used:

      Let :math:`x_\star` be a minimizer with :math:`F_\star = f(x_\star)`.
      For any :math:`x_{r_1} \in \calH` with :math:`g_{r_1} = \nabla f(x_{r_1})` and
      :math:`F_{r_1} = f(x_{r_1})`,

      .. math::
          F_{r_1} - F_\star \le \frac{1}{2\mu_{\textup{gd}}}\|g_{r_1}\|^2
          \quad \text{and} \quad
          F_{r_1} \ge F_\star.

    - Matrix/vector form used in :doc:`Interpolation conditions </theory/interpolation_conditions>`:

      With :math:`F=(F_{r_1},F_\star)` and
      :math:`z=(x_{r_1},x_\star,g_{r_1},0)`, the two inequalities are
      encoded by

      .. math::
          a_1 = (-1, 1), \qquad
          M_1 =
          \begin{bmatrix}
              0 & 0 & 0 & 0 \\
              0 & 0 & 0 & 0 \\
              0 & 0 & 0 & 0 \\
              0 & 0 & 0 & 0
          \end{bmatrix},

      and

      .. math::
          a_2 = (1, -1), \qquad
          M_2 =
          \begin{bmatrix}
              0 & 0 & 0 & 0 \\
              0 & 0 & 0 & 0 \\
              0 & 0 & -\frac{1}{2\mu_{\textup{gd}}} & 0 \\
              0 & 0 & 0 & 0
          \end{bmatrix}.

    **Parameters**

    - `mu_gd` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The gradient-dominated parameter corresponding
      to :math:`\mu_{\textup{gd}}` (must be :math:`> 0` and finite).

    **Raises**

    - `ValueError`: If `mu_gd` is not a number, :math:`\le 0`, or infinite.

    **Note**

    - This condition is only sufficient for AutoLyap analyses; tightness of
      the resulting bound is not guaranteed.
    - When used inside :class:`~autolyap.problemclass.InclusionProblem`,
      the problem must have exactly one component. In the notation of
      :doc:`3. Algorithm representation </theory/algorithm_representation>`,
      :math:`m = 1`, :math:`m_{\textup{func}} = 1`, and
      :math:`m_{\textup{op}} = 0`. This single component may still
      contain a list of function conditions (an intersection).
    """
    def __init__(self, mu_gd: Union[int, float]) -> None:
        mu_gd = _ensure_positive_finite(
            mu_gd,
            "Gradient-dominated parameter",
            "Gradient-dominated parameter (mu_gd) must be greater than 0 and finite.",
        )
        self.mu_gd = mu_gd

    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, _InterpolationIndices]]:
        r"""
        Return interpolation data for gradient-dominated functions.

        The tuple format and notation follow the class-level reference in
        :class:`~autolyap.problemclass.GradientDominated`.

        **Returns**

        - (:class:`~typing.List`\[:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`numpy.ndarray`, :class:`bool`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]\]): A list containing two
          tuples with the interpolation data.

        """
        a1 = np.array([-1, 1])
        M1 = np.zeros((4, 4))
        
        a2 = np.array([1, -1])
        M2 = np.zeros((4, 4))
        M2[2, 2] = -1 / (2 * self.mu_gd)
        
        interp_idx = _InterpolationIndices("r1!=star")
        eq_flag = False
        return [
            (M1, a1, eq_flag, interp_idx),
            (M2, a2, eq_flag, interp_idx)
        ]



"""Interpolation index helpers for problem classes."""

class _InterpolationIndices:
    r"""
    Wrapper for interpolation indices that enforces allowed values.

    Class-level reference
    =====================

    This class-level docstring centralizes the allowed symbolic labels used to
    index interpolation pairs in problem classes.

    **Allowed values**

    - ``r1<r2``: unordered pairs of distinct interpolation entries :math:`r_1` and :math:`r_2`.
    - ``r1!=r2``: ordered pairs of distinct interpolation entries :math:`r_1` and :math:`r_2`.
    - ``r1``: single-entry interpolation constraints (index :math:`r_1`).
    - ``r1!=star``: pairs between any interpolation entry :math:`r_1` and the distinguished star entry.

    **Parameters**

    - `value` (:class:`str`): The interpolation index string.

    **Raises**

    - `ValueError`: If `value` is not in the allowed set.
    """
    # Frozen set prevents accidental runtime mutation of allowed values.



    ALLOWED_VALUES = frozenset({"r1<r2", "r1!=r2", "r1", "r1!=star"})
    __slots__ = ("_value",)

    def __init__(self, value: str) -> None:
        self.value = value

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError("Interpolation index must be a string.")
        if value not in self.ALLOWED_VALUES:
            raise ValueError(f"Invalid interpolation index: {value}. Allowed values: {self.ALLOWED_VALUES}")
        self._value = value

    def __str__(self) -> str:
        return self._value

    def __eq__(self, other: object) -> bool:
        if isinstance(other, _InterpolationIndices):
            return self._value == other._value
        if isinstance(other, str):
            return self._value == other
        return False

    def __repr__(self) -> str:
        return f"_InterpolationIndices({self._value})"



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class NesterovConstant(Algorithm):
    r"""
    Nesterov's constant-step scheme :cite:`nesterov2018lecturesconvexoptimization{Constant step scheme III}`.

    See :doc:`3. Algorithm representation </theory/algorithm_representation>`
    for mathematical notation and definitions.

    Notation-driven assumptions are declared by the user via
    :class:`~autolyap.problemclass.InclusionProblem`: when present, terms written with
    :math:`\nabla` use differentiable functions, terms written with
    :math:`\prox_{\gamma f}` use proper, lower semicontinuous, convex functions,
    and terms written with :math:`J_{\gamma G}` use maximally monotone operators.

    Standard form
    -------------


    Let :math:`q = \mu/L` and
    :math:`\eta = (1-\sqrt{q})/(1+\sqrt{q})`. For initial points
    :math:`x^{-1},x^0 \in \calH`,

    .. math::
        (\forall k \in \naturals)\quad
        \left[
        \begin{aligned}
            y^k &= x^k + \eta (x^k - x^{k-1}), \\
            x^{k+1} &= y^k - \frac{1}{L}\nabla f(y^k).
        \end{aligned}
        \right.

    State-space representation
    --------------------------

    The update can be written in the algorithm representation with

    .. math::
        \bx^k = (x^k, x^{k-1}), \qquad
        \bu^k = \nabla f(y^k), \qquad
        \by^k = y^k.

    Let :math:`q = \mu/L`. With this representation, the system matrices are

    .. math::
        \begin{aligned}
            A_k &=
            \begin{bmatrix}
            \frac{2}{1+\sqrt{q}} & -\frac{1-\sqrt{q}}{1+\sqrt{q}} \\
            1 & 0
            \end{bmatrix}, &
            B_k &=
            \begin{bmatrix}
            -\frac{1}{L} \\
            0
            \end{bmatrix}, \\
            C_k &=
            \begin{bmatrix}
            \frac{2}{1+\sqrt{q}} & -\frac{1-\sqrt{q}}{1+\sqrt{q}}
            \end{bmatrix}, &
            D_k &= \begin{bmatrix} 0 \end{bmatrix}.
        \end{aligned}

    These are the system matrices returned by :meth:`~autolyap.algorithms.Algorithm.get_ABCD`.

    Structural parameters
    ---------------------

    .. math::
        n = 2,\quad m = 1,\quad (\bar{m}_i)_{i=1}^{m} = (1),\quad \bar{m} = 1.

    .. math::
        I_{\text{func}} = \{1\},\quad I_{\text{op}} = \varnothing.
    """
    def __init__(self, mu: float, L: float) -> None:
        r"""
        Initialize the constant-step scheme.

        Structural inputs passed to :class:`~autolyap.algorithms.Algorithm` are

        .. math::
            n = 2,\quad m = 1,\quad (\bar m_i)_{i=1}^{m} = (1),\quad \bar m = 1,\quad
            I_{\mathrm{func}} = \{1\},\quad I_{\mathrm{op}} = \varnothing.
        """
        super().__init__(2, 1, [1], [1], [])
        self.set_L(L)
        self.set_mu(mu)

    def set_mu(self, mu: float) -> None:
        r"""
        Set the strong-convexity parameter :math:`\mu`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.NesterovConstant`.

        **Parameters**

        - `mu` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`\mu`.

        **Raises**

        - `ValueError`: If `mu` is not a finite real number or if :math:`\mu \le 0`.
        """
        mu = self._validate_positive_finite_real(mu, "mu")
        self._set_dynamic_parameter("mu", mu)

    def set_L(self, L: float) -> None:
        r"""
        Set the smoothness parameter :math:`L`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.NesterovConstant`.

        **Parameters**

        - `L` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`L`.

        **Raises**

        - `ValueError`: If `L` is not a finite real number or if :math:`L \le 0`.
        """
        L = self._validate_positive_finite_real(L, "L")
        self._set_dynamic_parameter("L", L)
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        q = self.mu / self.L
        
        A = np.array([[2/(1+np.sqrt(q)),-(1-np.sqrt(q))/(1+np.sqrt(q))],[1,0]])
        B = np.array([[-1 / self.L], [0]])
        C = np.array([[2/(1+np.sqrt(q)),-(1-np.sqrt(q))/(1+np.sqrt(q))]])
        D = np.array([[0]])
        return (A, B, C, D)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm


class MalitskyTamFRB(Algorithm):
    r"""
    Malitsky--Tam forward-reflected-backward method :cite:`malitsky2020forwardbackwardsplitting`.

    See :doc:`3. Algorithm representation </theory/algorithm_representation>`
    for mathematical notation and definitions.

    Notation-driven assumptions are declared by the user via
    :class:`~autolyap.problemclass.InclusionProblem`: when present, terms written with
    :math:`\nabla` use differentiable functions, terms written with
    :math:`\prox_{\gamma f}` use proper, lower semicontinuous, convex functions,
    and terms written with :math:`J_{\gamma G}` use maximally monotone operators.

    Standard form
    -------------

    For an initial pair :math:`(x^{-1}, x^0) \in \calH^2` and step size
    :math:`\gamma \in \reals_{++}`,

    .. math::
        (\forall k \in \naturals)\quad
        x^{k+1} = J_{\gamma G_2}\left(x^k - 2\gamma G_1(x^k) + \gamma G_1(x^{k-1})\right).

    Equivalently,

    .. math::
        (\forall k \in \naturals)\quad
        x^{k+1} = x^k - 2\gamma G_1(x^k) + \gamma G_1(x^{k-1}) - \gamma G_2(x^{k+1}).

    State-space representation
    --------------------------

    The update can be written in the algorithm representation with

    .. math::
        \bx^k = (x^k,\; x^{k-1}), \qquad
        \bu^k = \left(G_1(x^k),\; G_1(x^{k-1}),\; G_2(x^{k+1})\right), \qquad
        \by^k = \left(x^k,\; x^{k-1},\; x^{k+1}\right).

    With this representation, the system matrices are

    .. math::
        \begin{aligned}
            A_k &=
            \begin{bmatrix}
            1 & 0 \\
            1 & 0
            \end{bmatrix}, &
            B_k &=
            \begin{bmatrix}
            -2\gamma & \gamma & -\gamma \\
            0 & 0 & 0
            \end{bmatrix}, \\
            C_k &=
            \begin{bmatrix}
            1 & 0 \\
            0 & 1 \\
            1 & 0
            \end{bmatrix}, &
            D_k &=
            \begin{bmatrix}
            0 & 0 & 0 \\
            0 & 0 & 0 \\
            -2\gamma & \gamma & -\gamma
            \end{bmatrix}.
        \end{aligned}

    These are the system matrices returned by :meth:`~autolyap.algorithms.Algorithm.get_ABCD`.

    Structural parameters
    ---------------------

    .. math::
        n = 2,\quad m = 2,\quad (\bar{m}_i)_{i=1}^{m} = (2,1),\quad \bar{m} = 3.

    .. math::
        I_{\text{func}} = \varnothing,\quad I_{\text{op}} = \{1,2\}.
    """
    def __init__(self, gamma: float) -> None:
        r"""
        Initialize the Malitsky--Tam forward-reflected-backward method.

        Structural inputs passed to :class:`~autolyap.algorithms.Algorithm` are

        .. math::
            n = 2,\quad m = 2,\quad (\bar m_i)_{i=1}^{m} = (2,1),\quad \bar m = 3,\quad
            I_{\mathrm{func}} = \varnothing,\quad I_{\mathrm{op}} = \{1,2\}.
        """
        super().__init__(2, 2, [2, 1], [], [1, 2])
        self.gamma = gamma

    def set_gamma(self, gamma: float) -> None:
        r"""
        Set the step-size parameter :math:`\gamma`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.MalitskyTamFRB`.

        **Parameters**

        - `gamma` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`\gamma`.

        **Raises**

        - `ValueError`: If `gamma` is not a finite real number or if :math:`\gamma \le 0`.
        """
        gamma = self._validate_positive_finite_real(gamma, "gamma")
        self._set_dynamic_parameter("gamma", gamma)

    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1, 0],
                      [1, 0]])
        B = np.array([[-2 * self.gamma, self.gamma, -self.gamma],
                      [0, 0, 0]])
        C = np.array([[1, 0],
                      [0, 1],
                      [1, 0]])
        D = np.array([[0, 0, 0],
                      [0, 0, 0],
                      [-2 * self.gamma, self.gamma, -self.gamma]])
        return (A, B, C, D)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class AcceleratedProximalPoint(Algorithm):
    r"""
    Accelerated proximal point method :cite:`kim2021acceleratedproximalpoint`.

    See :doc:`3. Algorithm representation </theory/algorithm_representation>`
    for mathematical notation and definitions.

    Notation-driven assumptions are declared by the user via
    :class:`~autolyap.problemclass.InclusionProblem`: when present, terms written with
    :math:`\nabla` use differentiable functions, terms written with
    :math:`\prox_{\gamma f}` use proper, lower semicontinuous, convex functions,
    and terms written with :math:`J_{\gamma G}` use maximally monotone operators.

    Standard form
    -------------


    Let :math:`\lambda_k = k/(k+2)`. For initial points
    :math:`x^0,y^0,y^{-1} \in \calH` and :math:`\gamma \in \reals_{++}`,

    .. math::
        (\forall k \in \naturals)\quad
        \left[
        \begin{aligned}
            x^{k+1} &=
            \begin{cases}
                J_{\gamma G}(y^k), & \text{if type = operator}, \\
                \prox_{\gamma f}(y^k), & \text{if type = function},
            \end{cases} \\
            y^{k+1} &= x^{k+1} + \lambda_k(x^{k+1} - x^k) - \lambda_k(x^k - y^{k-1}).
        \end{aligned}
        \right.

    State-space representation
    --------------------------

    The update can be written in the algorithm representation with

    .. math::
        \bx^k = (x^k, y^k, y^{k-1}), \qquad
        \bu^k = \frac{y^k - x^{k+1}}{\gamma}, \qquad
        \by^k = y^k.

    With :math:`\lambda_k = k/(k+2)`, the system matrices are

    .. math::
        \begin{aligned}
            A_k &=
            \begin{bmatrix}
            0 & 1 & 0 \\
            -2\lambda_k & 1+\lambda_k & \lambda_k \\
            0 & 1 & 0
            \end{bmatrix}, &
            B_k &=
            \begin{bmatrix}
            -\gamma \\
            -\gamma(1+\lambda_k) \\
            0
            \end{bmatrix}, \\
            C_k &= \begin{bmatrix} 0 & 1 & 0 \end{bmatrix}, &
            D_k &= \begin{bmatrix} -\gamma \end{bmatrix}.
        \end{aligned}

    These are the system matrices returned by :meth:`~autolyap.algorithms.Algorithm.get_ABCD`.

    Structural parameters
    ---------------------

    .. math::
        \text{type}=\text{"operator"}:\quad
        n = 3,\quad m = 1,\quad (\bar{m}_i)_{i=1}^{m} = (1),\quad \bar{m} = 1,\quad
        I_{\text{func}} = \varnothing,\quad I_{\text{op}} = \{1\}.

    .. math::
        \text{type}=\text{"function"}:\quad
        n = 3,\quad m = 1,\quad (\bar{m}_i)_{i=1}^{m} = (1),\quad \bar{m} = 1,\quad
        I_{\text{func}} = \{1\},\quad I_{\text{op}} = \varnothing.
    """
    def __init__(self, gamma: float, type: str = "operator") -> None:
        r"""
        Initialize the accelerated proximal point method.

        Structural inputs passed to :class:`~autolyap.algorithms.Algorithm` are
        case-dependent:

        - If `type = operator`:

          .. math::
              n = 3,\quad m = 1,\quad (\bar m_i)_{i=1}^{m} = (1),\quad \bar m = 1,\quad
              I_{\mathrm{func}} = \varnothing,\quad I_{\mathrm{op}} = \{1\}.

        - If `type = function`:

          .. math::
              n = 3,\quad m = 1,\quad (\bar m_i)_{i=1}^{m} = (1),\quad \bar m = 1,\quad
              I_{\mathrm{func}} = \{1\},\quad I_{\mathrm{op}} = \varnothing.
        """
        if type == "operator":
            super().__init__(3, 1, [1], [], [1])
        elif type == "function":
            super().__init__(3, 1, [1], [1], [])
        else:
            raise ValueError("type must be either 'operator' or 'function'")
        self.gamma = gamma
    
    def set_gamma(self, gamma: float) -> None:
        r"""
        Set the step-size parameter :math:`\gamma`.

        Shared notation follows the class-level reference in
        :class:`~autolyap.algorithms.AcceleratedProximalPoint`.

        **Parameters**

        - `gamma` (:class:`~typing.Union`\[:class:`int`, :class:`float`\]): The value corresponding to :math:`\gamma`.

        **Raises**

        - `ValueError`: If `gamma` is not a finite real number or if :math:`\gamma \le 0`.
        """
        gamma = self._validate_positive_finite_real(gamma, "gamma")
        self._set_dynamic_parameter("gamma", gamma)
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        lambda_var = k / (k + 2)
        A = np.array([[0, 1 , 0], 
                      [-2*lambda_var, 1+lambda_var, lambda_var],
                      [0, 1, 0]])
        B = np.array([[-self.gamma],
                      [-self.gamma*(1+lambda_var)],
                      [0]])
        C = np.array([[0, 1, 0]])
        D = np.array([[-self.gamma]])
        return (A, B, C, D)



"""Abstract interpolation-condition interfaces."""

import numpy as np
from abc import ABC, abstractmethod
from typing import List, Sequence, Tuple, Union

from autolyap.problemclass.indices import _InterpolationIndices

OperatorInterpolationData = Tuple[np.ndarray, _InterpolationIndices]
FunctionInterpolationData = Tuple[np.ndarray, np.ndarray, bool, _InterpolationIndices]
InterpolationData = Union[OperatorInterpolationData, FunctionInterpolationData]

class _InterpolationCondition(ABC):
    r"""
    Abstract base class for an interpolation condition.

    Class-level reference
    =====================

    This class-level docstring centralizes the data contract implemented by
    interpolation-condition subclasses.

    Derived classes must implement :meth:`get_data`, which returns interpolation data as a list
    of tuples encoding the relevant vectors, matrices, and interpolation indices.
    See :class:`~autolyap.problemclass.base._OperatorInterpolationCondition` and
    :class:`~autolyap.problemclass.base._FunctionInterpolationCondition` for
    the detailed notation and the corresponding interpolation constraints.

    """
    @abstractmethod
    def get_data(self) -> Sequence[InterpolationData]:
        r"""
        Return interpolation data.

        Shared tuple conventions follow the class-level reference in
        :class:`~autolyap.problemclass.base._InterpolationCondition`.

        **Returns**

        - (:class:`~typing.List`\[:class:`~typing.Union`\[:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\], :class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`numpy.ndarray`, :class:`bool`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]\]\]):
          A list of tuples representing the interpolation data.

        """
        pass

class _OperatorInterpolationCondition(_InterpolationCondition):
    r"""
    Base class for operator interpolation conditions.

    Class-level reference
    =====================

    This class-level docstring centralizes notation and tuple conventions for
    operator interpolation constraints.

    Return a list of tuples of the form:
    
    `(matrix, interpolation_indices)`

    where:

    - **matrix** is a square, symmetric 2D numpy array.
    - **interpolation_indices** is an instance of :class:`~autolyap.problemclass.indices._InterpolationIndices`.

    The returned **matrix** corresponds to the quadratic term
    :math:`M_{(i,o)}^{\textup{op}}`, where :math:`i \in \IndexOp` indexes the operator component
    and :math:`o` selects one of its interpolation constraints. These enter inequalities of the form

    .. math::
        \langle z, (M_{(i,o)}^{\textup{op}} \kron \Id) z \rangle \le 0,

    where :math:`z` stacks the relevant :math:`y_{i,j}^{k}` and :math:`u_{i,j}^{k}` entries selected
    by the interpolation indices. Concretely, if the indices select pairs
    :math:`(j_1,k_1),\ldots,(j_b,k_b)`, then

    .. math::
        z = (y_{i,j_1}^{k_1},\ldots,y_{i,j_b}^{k_b},u_{i,j_1}^{k_1},\ldots,u_{i,j_b}^{k_b}).

    For operator components, :math:`u_{i,j}^{k} \in G_i(y_{i,j}^{k})`.
    No linear term appears for operator conditions.

    For any matrix :math:`M \in \mathbb{R}^{d \times b}`, the tensor product
    :math:`M \kron \Id` denotes the linear map :math:`\calH^{b} \to \calH^{d}` defined by

    .. math::
        (M \kron \Id)z = \Big(\sum_{\ell=1}^{b}[M]_{1,\ell} z_\ell,\ldots,\sum_{\ell=1}^{b}[M]_{d,\ell} z_\ell\Big),

    for :math:`z = (z_1,\ldots,z_b) \in \calH^{b}`.

    """
    @abstractmethod
    def get_data(self) -> Sequence[OperatorInterpolationData]:
        r"""
        Return operator interpolation data.

        The tuple format and notation follow the class-level reference in
        :class:`~autolyap.problemclass.base._OperatorInterpolationCondition`.

        **Returns**

        - (:class:`~typing.List`\[:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]\]): A list of tuples, each containing a square
          symmetric matrix and an instance of :class:`~autolyap.problemclass.indices._InterpolationIndices`.

        """
        pass

class _FunctionInterpolationCondition(_InterpolationCondition):
    r"""
    Base class for function interpolation conditions.

    Class-level reference
    =====================

    This class-level docstring centralizes notation and tuple conventions for
    function interpolation constraints.

    Return a list of tuples of the form:
    
    `(matrix, vector, eq, interpolation_indices)`

    where:

    - **matrix** is a square, symmetric 2D numpy array.
    - **vector** is a 1D numpy array.
    - **eq** is a boolean flag (True for equality, False for inequality).
    - **interpolation_indices** is an instance of :class:`~autolyap.problemclass.indices._InterpolationIndices`.
    - **matrix** has shape :math:`(2b, 2b)` where :math:`b` is the length of **vector**.

    The returned **vector** corresponds to the linear term
    :math:`a_{(i,o)}^{\textup{func-ineq}}` or :math:`a_{(i,o)}^{\textup{func-eq}}`, where
    :math:`i \in \IndexFunc` indexes the functional component and :math:`o` selects one of its
    interpolation constraints, in function values,
    and the **matrix** corresponds to the quadratic term
    :math:`M_{(i,o)}^{\textup{func-ineq}}` or :math:`M_{(i,o)}^{\textup{func-eq}}` in constraints of the form

    .. math::
        \big(a_{(i,o)}^{\textup{func-ineq}}\big)^{\top}\begin{bmatrix}
            F_{i,j_1}^{k_1} \\
            \vdots \\
            F_{i,j_b}^{k_b}
        \end{bmatrix}
        + \langle z, (M_{(i,o)}^{\textup{func-ineq}} \kron \Id) z \rangle \le 0,

    and

    .. math::
        \big(a_{(i,o)}^{\textup{func-eq}}\big)^{\top}\begin{bmatrix}
            F_{i,j_1}^{k_1} \\
            \vdots \\
            F_{i,j_b}^{k_b}
        \end{bmatrix}
        + \langle z, (M_{(i,o)}^{\textup{func-eq}} \kron \Id) z \rangle = 0.

    Here :math:`z` stacks the corresponding :math:`y_{i,j_\ell}^{k_\ell}` and :math:`u_{i,j_\ell}^{k_\ell}`
    terms. Concretely, if the indices select pairs :math:`(j_1,k_1),\ldots,(j_b,k_b)`, then

    .. math::
        z = (y_{i,j_1}^{k_1},\ldots,y_{i,j_b}^{k_b},u_{i,j_1}^{k_1},\ldots,u_{i,j_b}^{k_b}).

    For functional components, :math:`u_{i,j}^{k} \in \partial f_i(y_{i,j}^{k})`, and
    :math:`F_{i,j}^{k} = f_i(y_{i,j}^{k})`. The flag **eq** selects equality vs. inequality.

    For any matrix :math:`M \in \mathbb{R}^{d \times b}`, the tensor product
    :math:`M \kron \Id` denotes the linear map :math:`\calH^{b} \to \calH^{d}` defined by

    .. math::
        (M \kron \Id)z = \Big(\sum_{\ell=1}^{b}[M]_{1,\ell} z_\ell,\ldots,\sum_{\ell=1}^{b}[M]_{d,\ell} z_\ell\Big),

    for :math:`z = (z_1,\ldots,z_b) \in \calH^{b}`.

    """
    @abstractmethod
    def get_data(self) -> Sequence[FunctionInterpolationData]:
        r"""
        Return function interpolation data.

        The tuple format and notation follow the class-level reference in
        :class:`~autolyap.problemclass.base._FunctionInterpolationCondition`.

        **Returns**

        - (:class:`~typing.List`\[:class:`~typing.Tuple`\[:class:`numpy.ndarray`, :class:`numpy.ndarray`, :class:`bool`, :class:`~autolyap.problemclass.indices._InterpolationIndices`\]\]): A list of tuples, each
          containing a square symmetric matrix, a 1D vector, a boolean flag, and an instance of
          :class:`~autolyap.problemclass.indices._InterpolationIndices`.

        """
        pass
